# -*- coding: utf-8 -*-
"""JWS (JSON Web Signature) module for the JOSE package."""

# Created: 2018-08-01 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import base64
from collections import OrderedDict
import json
from typing import Union, List, Optional

import nacl.hash
from sspyjose import Jose
from sspyjose import utils
from sspyjose.jwk import Jwk


class Jws(Jose):
    """
    JSON Web Signature (JWK) RFC 7515.

    Only implements the following cipher suites:

    - Ed25519 - public key-based, 256-bit key size (RFC 8037).
    """

    _header = None  # type: Optional[dict]
    """Protected (signed) header."""
    _payload = None  # type: Optional[dict]
    """Payload (signed)."""
    _signature = None  # type: Optional[bytes]
    """Signature."""
    _jwk = None  # type: Union[bytes, list[bytes], None]
    """Signing key."""
    _SERIALISATION_PARTS = ('header', 'payload', 'signature')
    """Parts of the JWS compact serialisation specification."""
    _DEFAULT_HEADER = OrderedDict()
    """Default header to use for this JWS type."""

    def __init__(self, *,
                 from_compact: Union[bytes, str, None] = None,
                 from_json: Union[bytes, str, None] = None,
                 jwk: Jwk = None):
        """
        Constructor.

        :param from_compact: If present, load data from compact ASCII JWS
            serialisation format.
        :param from_json: If present, load data from JSON JWS
            serialisation format. This form is to be used for
            multi-signature JWS objects.
        :param jwk: JWK to use for signing/validating.
        """
        super().__init__()
        self._header = self._DEFAULT_HEADER.copy()
        if from_compact:
            self.load_compact(from_compact)
        if from_json:
            self.load_json(from_json)
        if jwk:
            self._jwk = jwk

    @classmethod
    def get_instance(
            cls, *,
            alg: Optional[str] = None,
            from_compact: Union[bytes, str, None] = None,
            from_json: Union[bytes, str, None] = None,
            jwk: Optional[Jwk] = None) -> 'Jws':
        """
        Get an instance of a specific JWS object by algorithm via factory.

        :param alg: Specific algorithm indicator
            (default: {Jose.DEFAULT_SIG}).
        :param from_compact: Load data from compact ASCII JWS serialisation
            format (default: None).
        :param from_json: Load data from JSON JWS serialisation format
            (default: None). This form is to be used for multi-signature JWS
            objects.
        :param jwk: JWK to use for signing/verifying (default: None).
        :return: A {Jws} object.
        """
        if alg is None:
            alg = Jose.DEFAULT_SIG
        klass = _CLASS_MAPPING.get(alg)
        if klass:
            return klass(from_compact=from_compact, from_json=from_json,
                         jwk=jwk)
        raise RuntimeError('No suitable JWS type implemented for alg={}.'
                           .format(alg))

    @property
    def header(self) -> dict:
        """Protected header."""
        return self._header

    @header.setter
    def header(self, value: dict):
        """
        Set the protected header. This clears any signature attribute value.

        :param value: Content in bytes.
        """
        self._header = value
        self._signature = None

    @property
    def payload(self) -> dict:
        """Plain text message."""
        return self._payload

    @payload.setter
    def payload(self, value: dict):
        """
        Set payload to sign. This clears any signature attribute value.

        :param value: Content in bytes.
        """
        self._payload = value
        self._signature = None

    def _get_message(self) -> bytes:
        """Bytes of message and protected header to sign/validate."""
        header = utils.dict_to_base64(self._header).encode('utf-8')
        payload = utils.dict_to_base64(self._payload).encode('utf-8')
        return b'.'.join([header, payload])

    @property
    def signature(self) -> Union[bytes, List[bytes]]:
        """Signature."""
        return self._signature

    @signature.setter
    def signature(self, value: Union[bytes, List[bytes]]):
        """
        Set signature.

        :param value: Content in bytes.
        """
        self._signature = value

    @property
    def jwk(self) -> Jwk:
        """Signing/validating key for this JWS."""
        return self._jwk

    @jwk.setter
    def jwk(self, value: Jwk):
        """
        Set signing/validating key for this JWS.

        :param value: JWK key object.
        """
        self._jwk = value

    def add_jwk(self, value: Jwk):
        """
        Set signing/validating key for this JWS.

        :param value: JWK key object.
        """
        if self._jwk is None:
            self._jwk = value
        elif not isinstance(self._jwk, list):
            self._jwk = [self._jwk]
            self._jwk.append(value)
        else:
            self._jwk.append(value)

    def to_dict(self) -> dict:
        """Serialise to a dictionary for further string serialisation."""
        result = OrderedDict()
        if isinstance(self._signature, list):
            result['payload'] = utils.dict_to_base64(self._payload)
            result['signatures'] = []
            for key, signature in zip(self._jwk, self._signature):
                part = OrderedDict()
                part['protected'] = utils.dict_to_base64(self._header)
                header = {}
                if key.kid:
                    header['kid'] = key.kid
                else:
                    header['key_hash'] = nacl.hash.sha256(
                        key.x, encoder=nacl.encoding.RawEncoder)
                part['header'] = header
                part['signature'] = signature
                result['signatures'].append(part)
        else:
            for item in self._SERIALISATION_PARTS:
                part = getattr(self, '_' + item)
                if isinstance(part, dict):
                    part = utils.dict_to_base64(part)
                elif isinstance(part, bytes):
                    part = utils.bytes_to_string(part)
                elif part is None:
                    part = ''
                result[item] = part
        return result

    def to_json(self) -> str:
        """
        Serialise the JWS to JOSE compliant JSON.
        """
        return utils.dict_to_json(self.to_dict())

    def serialise(self, try_compact: bool = True) -> str:
        """
        Serialise the JWS.

        :param try_compact: Try the compact serialisation, if possible
            (default: True).
        :return: Serialised JWS.
        """
        if not try_compact or isinstance(self._signature, list):
            return self.to_json()
        base_serialised = self.to_dict()
        parts = [base_serialised[item]
                 for item in self._SERIALISATION_PARTS]
        return '.'.join(parts)

    @property
    def json(self) -> str:
        """Serialise JWS to JOSE compliant JSON."""
        return self.to_json()

    def load_compact(self, data: Union[bytes, str]):
        """
        Load JWS compact serialisation format.

        :param data: Serialisation data.
        """
        if isinstance(data, str):
            data = data.encode('utf-8')
        parts = data.split(b'.')
        if len(parts) != len(self._SERIALISATION_PARTS):
            raise ValueError(
                'Trying to load JWS compact serialisation with other than'
                ' {} parts: {} found.'
                .format(len(self._SERIALISATION_PARTS), len(parts)))
        parts = [base64.urlsafe_b64decode(item + b'==')
                 if item != b'' else None
                 for item in parts]
        content = dict(zip(self._SERIALISATION_PARTS, parts))
        content['header'] = json.loads(content['header'],
                                       object_pairs_hook=OrderedDict)
        content['payload'] = json.loads(content['payload'],
                                        object_pairs_hook=OrderedDict)
        for key, value in content.items():
            setattr(self, '_' + key, value)

    def load_json(self, data: Union[bytes, str]) -> List[dict]:
        """
        Load JWS JSON serialisation format.

        Note: Currently it is not possible to load multi-sig JWS with using
              different signing schemes/algorithms.

        :param data: Serialisation data.
        :return: List of the contained signatures' headers (containing e.g.
            the key IDs) in order of the signatures.
        """
        content = utils.json_to_dict(data)
        if set(content.keys()) != {'payload', 'signatures'}:
            raise ValueError('Got data incompatible with the form of JWS'
                             ' JSON serialisation for multi-signatures.')
        self._payload = utils.base64_to_dict(content['payload'])
        self._signature = []
        headers = []
        for signature in content['signatures']:
            headers.append(signature['header'])
            protected = utils.base64_to_dict(signature['protected'])
            if 'alg' not in self._header:
                self._header = protected
            elif protected['alg'] != self._header['alg']:
                raise RuntimeError(
                    'Object not suitable for loading a JWS {} signature'
                    ' into this {} object containing a {} signature.'
                    .format(protected['alg'], self.__class__,
                            self._header['alg']))
            # Need to 'binarise' the signature data, too
            self._signature.append(
                utils.string_to_bytes(signature['signature']))
        return headers

    def sign(self):
        """
        Sign the JWS payload and protected header.

        This also stores the resulting signature.
        """
        raise NotImplementedError('Needs to be implemented by child class.')

    def verify(self) -> bool:
        """
        Verify the JWS signature.

        :return: `True` on a valid signature.
        :raise: {nacl.exceptions.BadSignatureError} on an invalid signature.
        """
        raise NotImplementedError('Needs to be implemented by child class.')


class Ed25519Jws(Jws):
    """Ed25519 JWS signature (RFC 8037)."""

    _DEFAULT_HEADER = OrderedDict([('alg', 'Ed25519')])

    def sign(self):
        """
        Sign the JWS payload and protected header.

        This also stores the resulting signature.
        """
        if isinstance(self._jwk, list):
            self._signature = []
            for key in self._jwk:
                self._signature.append(
                    key._signing_key.sign(self._get_message()).signature)
        else:
            self._signature = (
                self._jwk._signing_key.sign(self._get_message()).signature)

    def verify(self) -> bool:
        """
        Verify the JWS signature.

        :return: `True` on a valid signature.
        :raise: {nacl.exceptions.BadSignatureError} on an invalid signature.
        """
        message = self._get_message()
        if (not isinstance(self._signature, list)
                and not isinstance(self._jwk, list)):
            self._jwk._verify_key.verify(message, signature=self._signature)
        else:
            no_keys = (1 if isinstance(self._jwk, Jwk)
                       else len(self._jwk))
            no_signatures = (1 if isinstance(self._signature, Jws)
                             else len(self._signature))
            if no_keys != no_signatures:
                raise nacl.exceptions.RuntimeError(
                    'Need {0} keys to verify {0} signatures, {1} given.'
                    .format(no_signatures, no_keys))
            message = self._get_message()
            for signature, jwk in zip(self._signature, self._jwk):
                jwk._verify_key.verify(message, signature=signature)
        return True


_CLASS_MAPPING = {
    Ed25519Jws._DEFAULT_HEADER['alg']: Ed25519Jws
}



# -*- coding: utf-8 -*-
"""JWK (JSON Web Key) module for the JOSE package."""

# Created: 2018-07-20 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

from collections import OrderedDict
import copy
import nacl.bindings
from typing import Iterable, Union, Optional

import nacl.public
import nacl.signing
from sspyjose import Jose
from sspyjose import cryptography_aead_wrap
from sspyjose import nacl_aead_wrap
from sspyjose import utils


class Jwk(Jose):
    """
    JSON Web Key (JWK) RFC 7517.

    Only implements key types for the following cipher suites:

    - ChaCha20/Poly1305 - 256-bit key size, 96-bit nonce
    - AES256-GCM - 256-bit key size, 96-bit nonce
    - X25519 - 256-bit key size
    - Ed25519 - 256-bit key (EdDSA with 256-bit key seed)

    Additional key information can be set within ``private`` data attributes
    (prefixed with an underscore ``_``), which will not be subject to
    serialisation.
    """

    _JSON_PRIVATE_KEY_ELEMENTS = None  # type: Optional[Iterable]
    """Elements of the key containing secrets."""
    _JSON_BINARY_ELEMENTS = None  # type: Optional[Iterable]
    """Elements of the key containing binary data."""

    def __init__(self, kty: str):
        """
        Constructor.

        :param kty: Key type.
        """
        super().__init__()
        self._data['kty'] = kty

    @classmethod
    def get_instance(
            cls, *,
            alg: Optional[str] = None,
            crv: Optional[str] = None,
            from_json: Optional[str] = None,
            from_dict: Optional[dict] = None,
            generate: bool = False,
            from_secret: Union[str, bytes, None] = None) -> 'Jwk':
        """
        Get an instance of a specific JWK object by algorithm via factory.

        :param alg: Specific encryption algorithm indicator (default: None).
        :param crv: Elliptic curve to use for the public key algorithm
            (default: None).
        :param from_json: Initialise the key from the given JSON
            representation of JWK key content (default: None).
        :param from_dict: Initialise the key from the given dictionary
            representation of JWK key content (equivalent content to JSON,
            default: None).
        :param generate: Generate a new key (default: False).
        :param from_secret: From the secret key (default: None). Only
            applicable to symmetric keys!
        :return: A {Jwk} object or `None` if no suitable class is available.
        """
        if alg is None and crv is None:
            if from_dict:
                alg = from_dict.get('alg')
                crv = from_dict.get('crv')
            elif from_json:
                content = utils.json_to_dict(from_json)
                alg = content.get('alg')
                crv = content.get('crv')
        if alg is None and crv is None:
            # Still nothing, set the default.
            alg = Jose.DEFAULT_ENC
        klass = _CLASS_MAPPING.get((alg, crv))
        if klass:
            if (from_secret
                    and alg in [ChaCha20Poly1305Jwk._DEFAULT_HEADER['alg'],
                                AES256GCMJwk._DEFAULT_HEADER['alg']]):
                return klass(from_json=from_json, from_dict=from_dict,
                             generate=generate, from_secret=from_secret)
            else:
                return klass(from_json=from_json, from_dict=from_dict,
                             generate=generate)
        raise RuntimeError(
            'No suitable JWK type implemented for alg={} and crv={}.'
            .format(alg, crv))

    @property
    def kty(self) -> str:
        """Key type."""
        return self._data['kty']

    @property
    def kid(self) -> str:
        """Key ID."""
        return self._data.get('kid')

    def to_dict(self, exclude_private: bool = False) -> dict:
        """
        Express the JWK as a JOSE compliant dictionary.

        :param exclude_private: Optional parameter to exclude private key
            information in the JSON serialisation.
        :return: Dictionary representation of key.
        """
        to_remove = (self._JSON_PRIVATE_KEY_ELEMENTS
                     if exclude_private else None)
        return utils.pack_bytes(self._data, to_remove)

    def to_json(self, exclude_private: bool = False) -> str:
        """
        Serialise the JWK to JOSE compliant JSON.

        :param exclude_private: Optional parameter to exclude private key
            information in the JSON serialisation.
        :return: JSON string of key.
        """
        to_remove = (self._JSON_PRIVATE_KEY_ELEMENTS
                     if exclude_private else None)
        return utils.dict_to_json(self._data, to_remove)

    @property
    def json(self) -> str:
        """Serialise JWK as JOSE compliant JSON."""
        return self.to_json()

    def generate_key(self):
        """Generate a new key pair for this JWK object."""
        raise NotImplementedError('Needs to be implemented by child class.')

    def from_json(self, content: str):
        """
        Load a new key from JSON content into this JWK object.

        :param content: JSON representation of JWK key content.
        """
        raise NotImplementedError('Needs to be implemented by child class.')

    def from_dict(self, content: dict):
        """
        Load a new key from JSON content into this JWK object.

        :param content: Dictionary representation of JWK key content.
        """
        raise NotImplementedError('Needs to be implemented by child class.')


class Ed25519Jwk(Jwk):
    """
    Edwards 25519 EdDSA signing key.
    """

    _JSON_PRIVATE_KEY_ELEMENTS = ('d',)
    _JSON_BINARY_ELEMENTS = ('x', 'd')
    _DEFAULT_HEADER = OrderedDict([('crv', 'Ed25519')])
    _signing_key = None  # type: Optional[nacl.signining.SigningKey]
    """PyNaCl signing key (private seed)."""
    _verify_key = None  # type: Optional[nacl.signining.VerifyKey]
    """PyNaCl verify key (public)."""

    def __init__(self, *,
                 from_json: Optional[str] = None,
                 from_dict: Optional[dict] = None,
                 generate: bool = False):
        """
        Constructor.

        :param from_json: Initialise the key from the given JSON
            representation of JWK key content (default: None).
        :param from_dict: Initialise the key from the given dictionary
            representation of JWK key content (equivalent content to JSON,
            default: None).
        :param generate: Generate a new key pair (default: False).
        """
        super().__init__('OKP')
        self._data.update(self._DEFAULT_HEADER)
        if from_json and generate:
            raise ValueError('Only one key generation method can be given.')
        if from_json:
            self.from_json(from_json)
        if from_dict:
            self.from_dict(from_dict)
        elif generate:
            self.generate_key()

    def generate_key(self):
        """Generate a new key pair for this JWK object."""
        signing_key = nacl.signing.SigningKey.generate()
        d = bytes(signing_key)
        x = bytes(signing_key.verify_key)
        self._data['d'] = d
        self._data['x'] = x
        self._data['_signing_key'] = signing_key

    def from_json(self, content: str):
        """
        Load a new key from JSON content into this JWK object.

        :param content: JSON representation of JWK key content.
        """
        self._data = utils.json_to_dict(
            content, binary=self._JSON_BINARY_ELEMENTS)

    def from_dict(self, content: dict):
        """
        Load a new key from JSON content into this JWK object.

        :param content: Dictionary representation of JWK key content.
        """
        content = copy.deepcopy(content)
        self._data = content
        if isinstance(self._data['x'], str):
            self._data['x'] = utils.string_to_bytes(content['x'])
        new_x = self._data['x']
        if 'd' in content and isinstance(content['d'], str):
            self.d = content['d']
            # Constant time hash comparison (do not use `!=` or `==`).
            if not nacl.bindings.sodium_memcmp(self._data['x'], new_x):
                raise RuntimeError(
                    'Public key `x` does not match `x` computed from'
                    ' private key `d`.')

    @property
    def crv(self) -> str:
        """Elliptic curve type."""
        return self._data['crv']

    @property
    def x(self) -> bytes:
        """Public key."""
        return self._data['x']

    @x.setter
    def x(self, value: Union[bytes, str]):
        """
        Set public key.

        :param value: Value of the public key. If a string is passed in, it
            will be decoded using URL safe base64 encoding.
        """
        pubkey = None
        if isinstance(value, bytes):
            pubkey = value
        elif isinstance(value, str):
            pubkey = utils.string_to_bytes(value)
        else:
            raise ValueError('Public key to set must be `bytes` or `str`.')
        if len(pubkey) != nacl.bindings.crypto_sign_PUBLICKEYBYTES:
            raise ValueError('Public key must be exactly {} bytes in size.'
                             .format(nacl.bindings.crypto_sign_PUBLICKEYBYTES))
        self._data['x'] = pubkey

    @property
    def d(self) -> bytes:
        """Private key seed."""
        return self._data.get('d')

    @d.setter
    def d(self, value: Union[bytes, str]):
        """
        Set signing key seed.

        :param value: Value of the key seed. If a string is passed in, it
            will be decoded using URL safe base64 encoding.
        """
        key_seed = None
        if isinstance(value, bytes):
            key_seed = value
        elif isinstance(value, str):
            key_seed = utils.string_to_bytes(value)
        else:
            raise ValueError('Key seed to set must be `bytes` or `str`.')
        if len(key_seed) != nacl.bindings.crypto_sign_SEEDBYTES:
            raise ValueError('Key seed must be exactly {} bytes in size.'
                             .format(nacl.bindings.crypto_sign_SEEDBYTES))
        self._data['d'] = key_seed
        signing_key = nacl.signing.SigningKey(key_seed)
        self._data['_signing_key'] = signing_key
        self._data['x'] = bytes(signing_key.verify_key)

    @property
    def _signing_key(self):
        """Signing key object for PyNaCl."""
        if '_signing_key' not in self._data:
            signing_key = nacl.signing.SigningKey(self._data['d'])
            self._data['_signing_key'] = signing_key
        return self._data['_signing_key']

    @property
    def _verify_key(self):
        """Verify key object for PyNaCl."""
        if '_verify_key' not in self._data:
            verify_key = nacl.signing.VerifyKey(self._data['x'])
            self._data['_verify_key'] = verify_key
        return self._data['_verify_key']


class X25519Jwk(Jwk):
    """X25519 Diffie-Hellman key."""

    _JSON_PRIVATE_KEY_ELEMENTS = ('d',)
    _JSON_BINARY_ELEMENTS = ('x', 'd')
    _DEFAULT_HEADER = OrderedDict([('crv', 'X25519')])
    _dh_key = None  # type: Optional[nacl.public.PrivateKey]
    """PyNaCl Diffie-Hellman key (private)."""

    def __init__(self, *,
                 from_json: Optional[str] = None,
                 from_dict: Optional[dict] = None,
                 generate: bool = False):
        """
        Constructor.

        :param from_json: Initialise the key from the given JSON
            representation of JWK key content (default: None).
        :param from_dict: Initialise the key from the given dictionary
            representation of JWK key content (equivalent content to JSON,
            default: None).
        :param generate: Generate a new key pair (default: False).
        """
        super().__init__('OKP')
        self._data.update(self._DEFAULT_HEADER)
        if from_json and generate:
            raise ValueError('Only one key generation method can be given.')
        if from_json:
            self.from_json(from_json)
        if from_dict:
            self.from_dict(from_dict)
        elif generate:
            self.generate_key()

    def generate_key(self):
        """Generate a new key pair for this JWK object."""
        my_key = nacl.public.PrivateKey.generate()
        d = bytes(my_key)
        x = bytes(my_key.public_key)
        self._data['d'] = bytes(d)
        self._data['_dh_key'] = my_key
        self._data['x'] = x

    def from_json(self, content: str):
        """
        Load a new key from JSON content into this JWK object.

        :param content: JSON representation of JWK key content.
        """
        self._data = utils.json_to_dict(
            content, binary=self._JSON_BINARY_ELEMENTS)

    def from_dict(self, content: dict):
        """
        Load a new key from JSON content into this JWK object.

        :param content: Dictionary representation of JWK key content.
        """
        content = copy.deepcopy(content)
        self._data = content
        if isinstance(self._data['x'], str):
            self._data['x'] = utils.string_to_bytes(content['x'])
        new_x = self._data['x']
        if 'd' in content and isinstance(content['d'], str):
            self.d = content['d']
            # Constant time hash comparison (do not use `!=` or `==`).
            if not nacl.bindings.sodium_memcmp(self._data['x'], new_x):
                raise RuntimeError(
                    'Public key `x` does not match `x` computed from'
                    ' private key `d`.')

    @property
    def crv(self) -> str:
        """Elliptic curve type."""
        return self._data['crv']

    @property
    def x(self) -> bytes:
        """Public key."""
        return self._data['x']

    @property
    def d(self) -> bytes:
        """Private key."""
        return self._data['d']

    @d.setter
    def d(self, value: Union[bytes, str]):
        """
        Set private key.

        :param value: Value of the private key. If a string is passed in, it
            will be decoded using URL safe base64 encoding.
        """
        private_key = None
        if isinstance(value, bytes):
            private_key = value
        elif isinstance(value, str):
            private_key = utils.string_to_bytes(value)
        else:
            raise ValueError('Private key to set must be `bytes` or `str`.')
        if len(private_key) != nacl.public.PrivateKey.SIZE:
            raise ValueError('Private key must be exactly {} bytes in size.'
                             .format(nacl.public.PrivateKey.SIZE))
        my_key = nacl.public.PrivateKey(private_key)
        self._data['d'] = private_key
        self._data['_dh_key'] = my_key
        self._data['x'] = bytes(my_key.public_key)

    @property
    def _dh_key(self):
        """Diffie-Hellman key object for PyNaCl."""
        if '_dh_key' not in self._data:
            if 'd' in self._data:
                dh_key = nacl.public.PrivateKey(self._data['d'])
            else:
                dh_key = nacl.public.PublicKey(self._data['x'])
            self._data['_dh_key'] = dh_key
        return self._data['_dh_key']


class SymmetricJwk(Jwk):
    """Symmetric key."""

    _JSON_PRIVATE_KEY_ELEMENTS = ('k',)
    _JSON_BINARY_ELEMENTS = ('k',)
    _DEFAULT_HEADER = None  # type: Optional[dict]
    _KEY_SIZE = None  # type: Optional[int]

    def __init__(self, *,
                 from_json: Optional[str] = None,
                 from_dict: Optional[dict] = None,
                 from_secret: Union[str, bytes, None] = None,
                 generate: bool = False):
        """
        Constructor.

        :param from_json: Initialise the key from the given JSON
            representation of JWK key content (default: None).
        :param from_dict: Initialise the key from the given dictionary
            representation of JWK key content (equivalent content to JSON,
            default: None).
        :param from_secret: From the secret key.
        :param generate: Generate a new key (default: False).
        """
        super().__init__('oct')
        self._data.update(self._DEFAULT_HEADER)
        if generate and (from_json or from_dict or from_secret):
            raise ValueError('Only one key generation method can be given.')
        if generate:
            self.generate_key()
        elif from_json:
            self.from_json(from_json)
        elif from_dict:
            self.from_dict(from_dict)
        elif from_secret:
            if isinstance(from_secret, str):
                from_secret = utils.string_to_bytes(from_secret)
            self._data['k'] = from_secret

    def generate_key(self):
        """Generate a new key for this JWK object."""
        self._data['k'] = nacl.utils.random(self._KEY_SIZE)

    def from_json(self, content: str):
        """
        Load a new key from JSON content into this JWK object.

        :param content: JSON representation of JWK key content.
        """
        self._data = utils.json_to_dict(
            content, binary=self._JSON_BINARY_ELEMENTS)

    def from_dict(self, content: dict):
        """
        Load a new key from JSON content into this JWK object.

        :param content: Dictionary representation of JWK key content.
        """
        content = copy.deepcopy(content)
        if isinstance(content['k'], str):
            content['k'] = utils.string_to_bytes(content['k'])
        self._data = content

    @property
    def k(self) -> bytes:
        """Encryption key."""
        return self._data['k']

    @k.setter
    def k(self, value: Union[bytes, str]):
        """
        Set encryption key.

        :param value: Value of the key. If a string is passed in, it will be
            decoded using URL safe base64 encoding.
        """
        key = None
        if isinstance(value, bytes):
            key = value
        elif isinstance(value, str):
            key = utils.string_to_bytes(value)
        else:
            raise ValueError('Key to set must be `bytes` or `str`.')
        if len(key) != self._KEY_SIZE:
            raise ValueError('Key must be exactly {} bytes in size.'
                             .format(self._KEY_SIZE))
        self._data['k'] = key

    @property
    def alg(self) -> str:
        """Encryption algorithm (`alg` attribute)."""
        return self._data['alg']

    @property
    def use(self) -> str:
        """Key usage."""
        return self._data['use']


class ChaCha20Poly1305Jwk(SymmetricJwk):
    """ChaCha20/Poly1305 symmetric key."""

    _DEFAULT_HEADER = OrderedDict([('alg', 'C20P'),
                                   ('use', 'enc')])
    _KEY_SIZE = nacl_aead_wrap.KEYBYTES


class AES256GCMJwk(SymmetricJwk):
    """AES256-GCM symmetric key."""

    _DEFAULT_HEADER = OrderedDict([('alg', 'A256GCM'),
                                   ('use', 'enc')])
    _KEY_SIZE = cryptography_aead_wrap.KEYBYTES


_CLASS_MAPPING = {
    (ChaCha20Poly1305Jwk._DEFAULT_HEADER['alg'], None): ChaCha20Poly1305Jwk,
    (AES256GCMJwk._DEFAULT_HEADER['alg'], None): AES256GCMJwk,
    (None, Ed25519Jwk._DEFAULT_HEADER['crv']): Ed25519Jwk,
    (None, X25519Jwk._DEFAULT_HEADER['crv']): X25519Jwk
}



# -*- coding: utf-8 -*-
"""Functional utilities."""

# Created: 2018-07-24 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import base64
from collections import OrderedDict
import json
import re
from typing import Iterable, Union, Optional

_NUMBER_TEST = re.compile(r'^[0-9]+$')


def bytes_to_string(value: bytes) -> str:
    """
    Convert `bytes` to a base64 URL encoded string.

    Trailing padding characters (`=`) are stripped.

    :param value: Bytes to be converted.
    :return: String representation.
    """
    return base64.urlsafe_b64encode(value).strip(b'=').decode('utf-8')


def string_to_bytes(value: str) -> bytes:
    """
    Convert base64 URL encoded string to `bytes`.

    Trailing padding characters (`=`) are added.

    :param value: String representation.
    :return: Bytes representation.
    """
    return base64.urlsafe_b64decode(value.encode('utf-8') + b'==')


def _transform_simple(data):
    """Transform a simple data type (not an iterable) for serialisation."""
    if isinstance(data, bytes):
        return bytes_to_string(data)
    else:
        return data


def pack_bytes(data: dict, to_remove: Optional[Iterable[str]] = None) -> dict:
    """
    Convert `bytes` in a dictionary to a copy with base64 URL strings.

    Objects elements with a key starting with an underscore ``_``
    are considered "private" are stripped out.

    :param data: JSON data as a dictionary.
    :param to_remove: Optionally object attributes to remove from
        serialisation (default: None).
    :return: Binary converted dictionary representation of JOSE data.
    """
    if not isinstance(data, dict):
        return _transform_simple(data)
    result = OrderedDict()
    to_remove = to_remove or []
    for key, value in data.items():
        if isinstance(key, str) and _NUMBER_TEST.match(key):
            key = int(key)
        if isinstance(key, str) and (key.startswith('_') or key in to_remove):
            continue
        elif isinstance(value, dict):
            result[key] = pack_bytes(value, to_remove)
        elif isinstance(value, (list, tuple)):
            result[key] = [pack_bytes(item, to_remove)
                           for item in value]
        else:
            result[key] = _transform_simple(value)
    return result


def unpack_bytes(data: dict, binary: Optional[Iterable[str]] = None) -> dict:
    """
    Convert base64 URL encoded strings in a JOSE dictionary to `bytes`.

    :param data: JSON data as a dictionary.
    :param binary: Optionally an iterable of strings for elements that
        must be converted from a base64 URL encoded representation
        into bytes.
    :return: Binary converted dictionary representation of JOSE data.
    """
    binary = binary or []
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = unpack_bytes(value, binary)
        elif isinstance(value, str) and key in binary:
            data[key] = string_to_bytes(value)
    return data


def dict_to_json(data: dict, to_remove: Optional[Iterable[str]] = None) -> str:
    """
    Convert a JOSE representation of a JOSE object to JSON representation.

    All binary data (bytes) will be base64 URL encoded. Objects elements
    with a key starting with an underscore ``_`` are considered "private"
    are stripped out.

    :param data: Python dictionary representation of JOSE data.
    :param to_remove: Optionally object attributes to remove from
        serialisation (default: None).
    :return: JSON encoded data.
    """
    return json.dumps(pack_bytes(data, to_remove),
                      indent=None, separators=(',', ':'))


def json_to_dict(data: Union[str, bytes],
                 binary: Optional[Iterable[str]] = None) -> dict:
    """
    Convert a JSON representation of a JOSE object to a Python dict.

    :param data: JSON encoded data.
    :param binary: Optionally an iterable of strings for elements that
        must be converted from a base64 URL encoded representation
        into bytes.
    :return: Python dictionary representation of JOSE data.
    """
    return unpack_bytes(json.loads(data, object_pairs_hook=OrderedDict),
                        binary)


def dict_to_base64(data: dict,
                   to_remove: Optional[Iterable[str]] = None) -> str:
    """
    Convert a JOSE dict to a URL-safe base64 representation of a JSON string.

    All binary data (bytes) will be base64 URL encoded. Objects elements
    with a key starting with an underscore ``_`` are considered "private"
    are stripped out.

    :param data: Python dictionary representation of JOSE data.
    :param to_remove: Optionally object attributes to remove from
        serialisation (default: None).
    :return: Encoded string representation.
    """
    json_form = json.dumps(pack_bytes(data, to_remove),
                           indent=None, separators=(',', ':'))
    return bytes_to_string(json_form.encode('utf-8'))


def base64_to_dict(data: str, binary: Optional[Iterable[str]] = None) -> dict:
    """
    Convert a URL base64 representation of a JSON string to a `dict`.

    :param data: Base64 string representation of JOSE data.
    :param binary: Optionally an iterable of strings for elements that
        must be converted from a base64 URL encoded representation
        into bytes.
    :return: Dictionary representation.
    """
    json_form = string_to_bytes(data)
    return json_to_dict(json_form, binary)



# -*- coding: utf-8 -*-
"""JSON Object signing and Encryption (JOSE, RFC 7520) base."""

# Created: 2018-07-25 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'


class Jose:
    """
    JSON Object signing and Encryption (JOSE, RFC 7520) base class.

    JOSE is split into 5 main components:

    - JSON Web Algorithms (JWA, RFC 7518).
    - JSON Web Encryption (JWE, RFC 7516).
    - JSON Web Key (JWK, RFC 7517).
    - JSON Web Signature (JWS, RFC 7515).
    - JSON Web Token (JWT, RFC 7519).

    Additional specifications and drafts implemented:

    - ChaCha20/Poly1305 JWE symmetric key encryption (RFC 7539).
    - Ed25519 JWS signature (RFC 8037).
    - JSON Web Key (JWK) Thumbprint (RFC 7638).
    - JWS Unencoded Payload Option (RFC 7797).

    We need JWK, JWE and JWS.
    """

    DEFAULT_ENC = 'C20P'
    """Default cipher for symmetric encryption."""
    DEFAULT_SIG = 'Ed25519'
    """Default cipher for cryptographic signatures."""
    _data = None  # type: dict
    """JOSE object data."""

    def __init__(self):
        """Constructor."""
        self._data = {}

    def to_json(self):  # noqa: D102
        raise NotImplementedError('Needs to be implemented by child class.')

    @staticmethod
    def from_json(content):  # noqa: D102
        raise NotImplementedError('Needs to be implemented by child class.')



# -*- coding: utf-8 -*-
"""JWE (JSON Web Encryption) module for the JOSE package."""

# Created: 2018-07-30 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

import base64
from collections import OrderedDict
import nacl.utils
from typing import Optional, Union, Callable

from sspyjose import Jose
from sspyjose import cryptography_aead_wrap as cryptography_aead
from sspyjose import nacl_aead_wrap as nacl_aead
from sspyjose import utils
from sspyjose.jwk import (Jwk,
                          X25519Jwk)


def _int_to_bytes(value: int) -> bytes:
    """
    Convert an unsigned integer to a 32 bit byte representation.

    The result is in big-endian format.

    :param value: The (unsigned) integer value to convert.
    :return: Byte representation.
    """
    return value.to_bytes(4, byteorder='big', signed=False)


class Jwe(Jose):
    """
    JSON Web Encryption (JWE) RFC 7516.

    Only implements the following cipher suites:

    - C20P - symmetric, 256-bit key size, 96-bit nonce
      [https://tools.ietf.org/id/draft-amringer-jose-chacha-00.txt]
      (RFC 7539 and RFC 8439).
    - A256GCM - symmetric, 256-bit key size, 96-bit nonce, no key wrapping
      (basic RFC 7516).
    - X25519 - public key-based, 256-bit key size (RFC 8037).
    - `unsecured` - non-standard, unencrypted, 'symmetric' JWE
      (contains an `{"alg": "unsecured"}` header).
    """

    _header = None  # type: Optional[dict]
    """Protected header."""
    _key = None  # type: Optional[bytes]
    """Encrypted key."""
    _nonce = None  # type: Optional[bytes]
    """Nonce/initialisation vector."""
    _message = None  # type: Optional[dict]
    """Plain text message."""
    _ciphertext = None  # type: Optional[bytes]
    """Ciphertext."""
    _tag = None  # type: Optional[bytes]
    """Authentication tag/MAC."""
    _SERIALISATION_PARTS = ('header', 'key', 'nonce', 'ciphertext', 'tag')
    """Parts of the JWE serialisation specification."""
    _jwk = None  # type: Optional[Jwk]
    """JWK to use for en/decryption."""
    _message_bytes = None  # type: Optional[bytes]
    """Retains the byte serialisation of the message content in verbatim."""

    def __init__(self, *,
                 from_compact: Union[bytes, str, None] = None,
                 jwk: Optional[Jwk] = None):
        """
        Constructor.

        :param from_compact: If present, load data from compact ASCII JWE
            serialisation format.
        :param jwk: JWK to use for en-/decryption.
        """
        super().__init__()
        if from_compact:
            self.load_compact(from_compact)
        if jwk:
            self._jwk = jwk

    @classmethod
    def get_instance(
            cls, *,
            enc: Optional[str] = None,
            alg: Optional[str] = None,
            from_compact: Union[bytes, str, None] = None,
            jwk: Optional[Jwk] = None) -> 'Jwe':
        """
        Get an instance of a specific JWE object by algorithm via factory.

        :param enc: Specific encryption indicator
            (default: {Jose.DEFAULT_ENC}).
        :param alg: Specific algorithm indicator (default: 'dir').
        :param from_compact: Load data from compact ASCII JWE serialisation
            format (default: None).
        :param jwk: JWK to use for en-/decryption (default: None).
        :return: A {Jwe} object.
        """
        if enc is None:
            # Try some places to determine the required value for `enc`.
            # First from given key.
            if jwk:
                enc = jwk._data.get('alg')
            # Still nothing, try from JWE content.
            if enc is None and from_compact:
                if isinstance(from_compact, bytes):
                    from_compact = from_compact.decode('utf-8')
                elements = from_compact.split('.')
                content = utils.base64_to_dict(elements[0])
                enc = content.get('enc')
                alg = content.get('alg')
            # Nothing, so go with the default.
            if enc is None and alg != 'unsecured':
                enc = Jose.DEFAULT_ENC
        alg = alg or 'dir'
        klass = _CLASS_MAPPING.get((enc, alg))
        if klass:
            return klass(from_compact=from_compact, jwk=jwk)
        raise RuntimeError(
            'No suitable JWE type implemented for enc={} and alg={}.'
            .format(enc, alg))

    @property
    def header(self) -> dict:
        """Protected header."""
        return self._header

    @property
    def key(self) -> bytes:
        """Encrypted key."""
        return self._key

    @property
    def nonce(self) -> bytes:
        """Nonce/initialisation vector."""
        return self._nonce

    @property
    def message(self) -> dict:
        """Plain text message."""
        return self._message

    @message.setter
    def message(self, value: dict):
        """
        Set plain text message. This clears any ciphertext attribute value.

        :param value: Content in bytes.
        """
        self._message = value
        self._message_bytes = None
        self._ciphertext = None
        self._tag = None

    @property
    def ciphertext(self) -> bytes:
        """Ciphertext."""
        return self._ciphertext

    @ciphertext.setter
    def ciphertext(self, value: bytes):
        """
        Set ciphertext. This clears any message attribute value.

        :param value: Content in bytes.
        """
        self._ciphertext = value
        self._message = None

    @property
    def tag(self) -> bytes:  # noqa: D401
        """Authentication tag/MAC."""
        return self._tag

    @tag.setter
    def tag(self, value: bytes):
        """
        Set authentication tag/MAC.

        :param value: Content in bytes.
        """
        self._tag = value

    @property
    def jwk(self) -> Jwk:
        """En/decryption key for this JWE."""
        return self._jwk

    @jwk.setter
    def jwk(self, value: Jwk):
        """
        Set en/decryption key for this JWE.

        :param value: JWK key object.
        """
        self._jwk = value

    def to_dict(self) -> dict:
        """
        Express the JWE as a JOSE compliant dictionary.
        """
        result = OrderedDict()
        for item in self._SERIALISATION_PARTS:
            result[item] = getattr(self, '_' + item)
        # JSONify the header.
        result['header'] = utils.dict_to_base64(result['header'])
        return result

    def to_json(self) -> str:
        """
        Serialise the JWE to JOSE compliant JSON.
        """
        return utils.dict_to_json(self.to_dict())

    def serialise(self) -> str:
        """
        Serialise the JWE to JOSE (compact) form.
        """
        parts = [getattr(self, '_' + item)
                 for item in self._SERIALISATION_PARTS]
        # JSONify the header.
        parts[0] = utils.dict_to_json(parts[0]).encode('utf-8')
        parts = [utils.bytes_to_string(item)
                 if item else '' for item in parts]
        return '.'.join(parts)

    @property
    def json(self) -> str:  # noqa: D401
        """Serialised JWE as JOSE compliant JSON."""
        return self.to_json()

    def load_compact(self, data: Union[bytes, str]):
        """
        Load JWE compact serialisation format.

        :param data: Serialisation data.
        """
        if isinstance(data, str):
            data = data.encode('utf-8')
        parts = data.split(b'.')
        if len(parts) != len(self._SERIALISATION_PARTS):
            raise ValueError(
                'Trying to load JWE compact serialisation with other than'
                ' {} parts: {} found.'
                .format(len(self._SERIALISATION_PARTS), len(parts)))
        parts = [base64.urlsafe_b64decode(item + b'==')
                 if item != b'' else None
                 for item in parts]
        content = dict(zip(self._SERIALISATION_PARTS, parts))
        content['header'] = utils.json_to_dict(content['header'])
        for key, value in content.items():
            setattr(self, '_' + key, value)

    def get_message_bytes(self) -> bytes:
        """
        Get canonical message representation as bytes.
        """
        if self._message_bytes is None:
            self._message_bytes = utils.dict_to_json(
                self._message).encode('utf-8')
        return self._message_bytes

    def encrypt(self) -> (bytes, bytes):
        """
        Encrypt the JWE message, and store the resulting ciphertext and tag.

        :return: Tuple (ciphertext, tag).
        """
        raise NotImplementedError('Needs to be implemented by child class.')

    def decrypt(self) -> bytes:
        """
        Decrypts the JWE ciphertext and stores the resulting message.

        :return: Decrypted message.
        """
        raise NotImplementedError('Needs to be implemented by child class.')


class SymmetricJwe(Jwe):
    """
    Symmetric key encryption.
    """

    _DEFAULT_HEADER = None  # type: Optional[dict]
    _KEYBYTES = 0  # type: Optional[int]
    _NONCEBYTES = 0  # type: Optional[int]
    _aead_encrypt_function = None  # type: Optional[Callable]
    _aead_decrypt_function = None  # type: Optional[Callable]

    def encrypt(self) -> (bytes, bytes):
        """
        Encrypt the JWE message, and store the resulting ciphertext and tag.

        :return: Tuple (ciphertext, tag).
        """
        plaintext = self.get_message_bytes()
        self._header = self._DEFAULT_HEADER.copy()
        header_bytes = utils.dict_to_base64(self._header).encode('utf-8')
        if self._header['alg'] != 'unsecured':
            self._nonce = nacl.utils.random(nacl_aead.NONCEBYTES)
        else:
            result = plaintext, None
        if self.__class__._aead_encrypt_function:
            result = self.__class__._aead_encrypt_function(
                plaintext, header_bytes, self._nonce, self._jwk.k)
        self._ciphertext, self._tag = result
        return result

    def decrypt(self, *, allow_unsecured: bool = False) -> bytes:
        """
        Decrypts the JWE ciphertext and stores the resulting message.

        :param allow_unsecured: Allows explicitly for access to non-standard,
            "unsecured" JWEs (default: forbidden).
        :return: Decrypted message.
        """
        header_bytes = utils.dict_to_base64(self._header).encode('utf-8')
        result = None
        if self.__class__._aead_decrypt_function:
            result = self.__class__._aead_decrypt_function(
                self._ciphertext, self._tag, header_bytes, self._nonce,
                self._jwk.k)
        elif self._header['alg'] == 'unsecured':
            if allow_unsecured is not True:
                raise RuntimeError('Access to unsecured (non-standard) JWE is'
                                   ' forbidden unless explicitly allowed.')
            if self._nonce or self._tag or self._key:
                raise RuntimeError('Unsecured (non-standard) JWEs must not'
                                   ' contain a key, tag or nonce.')

            result = self._ciphertext
        self._message_bytes = result
        self._message = utils.json_to_dict(result)
        return self._message


class ChaCha20Poly1305Jwe(SymmetricJwe):
    """
    ChaCha20/Poly1305 symmetric key encryption `C20P`.

    - https://tools.ietf.org/id/draft-amringer-jose-chacha-00.txt
    - RFC 7539 and RFC 8439.
    """

    _DEFAULT_HEADER = OrderedDict([('alg', 'dir'),
                                   ('enc', 'C20P')])
    _KEYBYTES = nacl_aead.KEYBYTES
    _NONCEBYTES = nacl_aead.NONCEBYTES
    _aead_encrypt_function = nacl_aead.chacha20poly1305_encrypt
    _aead_decrypt_function = nacl_aead.chacha20poly1305_decrypt


class AES256GCMJwe(SymmetricJwe):
    """
    AES-256-GCM symmetric key encryption `A256GCM`.
    """

    _DEFAULT_HEADER = OrderedDict([('alg', 'dir'),
                                   ('enc', 'A256GCM')])
    _KEYBYTES = cryptography_aead.KEYBYTES
    _NONCEBYTES = cryptography_aead.NONCEBYTES
    _aead_encrypt_function = cryptography_aead.aes256gcm_encrypt
    _aead_decrypt_function = cryptography_aead.aes256gcm_decrypt


class UnsecuredJwe(SymmetricJwe):
    """
    Non-standard JWE object that is unsecured (not encrypted).
    """

    _DEFAULT_HEADER = OrderedDict([('alg', 'unsecured')])
    _KEYBYTES = None
    _NONCEBYTES = None
    _aead_encrypt_function = None
    _aead_decrypt_function = None


class X25519Jwe(Jwe):
    """
    X25519 public key encryption base class.

    Creating an RFC 7518 ECDH-ES key derivation of a suitable key as in RFC
    8037. This is a base class, that requires implementation of the payload
    encryption using a symmetric cipher.

    - RFC 7539
    """

    _DEFAULT_HEADER = None  # type: Optional[dict]
    _KEYBYTES = 0  # type: Optional[int]
    _NONCEBYTES = 0  # type: Optional[int]
    _aead_encrypt_function = None  # type: Optional[Callable]
    _aead_decrypt_function = None  # type: Optional[Callable]

    def ecdh_es_key(self, full_key: X25519Jwk, public_key: X25519Jwk,
                    *, apu: bytes = b'', apv: bytes = b'') -> bytes:
        """
        Compute and return the shared key for a symmetric JWE encryption.

        The payload is encrypted. This requires ephemeral sender and
        recipient JWK keys.

        Thi KDF uses the shared DH secret and various 'input parameters' as
        described in RFC 7518 (Sect. 4.6 and Appendix C), implementing the
        NIST Special Publication 800-56A, Revision 2 (May 2013, Sect. 5.8):

        - https://tools.ietf.org/html/rfc7518
        - http://dx.doi.org/10.6028/NIST.SP.800-56Ar2

        :param full_key: First (private) JWK.
        :param public_key: Second (public) JWK.
        :param apu: Agreement PartyUInfo is producer information
            (optional, default: b'').
        :param apv: Agreement PartyVInfo is recipient information
            (optional, default: b'').
        :return: Binary encryption key of required length.
        """
        # Prepare data needed for key derivation function (KDF).
        public_dh_key = public_key._dh_key
        if isinstance(public_dh_key, nacl.public.PrivateKey):
            public_dh_key = public_dh_key.public_key
        a_box = nacl.public.Box(full_key._dh_key, public_dh_key)
        shared_secret = a_box.shared_key()
        algorithm_id = self._DEFAULT_HEADER['enc'].encode('utf-8')
        algorithm_id_length = _int_to_bytes(len(algorithm_id))
        apu_length = _int_to_bytes(len(apu))
        apv_length = _int_to_bytes(len(apv))
        keydata_length = _int_to_bytes(self._KEYBYTES << 3)  # In bits.
        # Assemble data and derive key via hashing-based KDF.
        other_info = (algorithm_id_length + algorithm_id
                      + apu_length + apu
                      + apv_length + apv
                      + keydata_length)
        round_number = _int_to_bytes(1)  # 1st (and only) round of hashing.
        hash_input = round_number + shared_secret + other_info
        derived_secret = nacl.hash.sha256(hash_input,
                                          encoder=nacl.encoding.RawEncoder)
        derived_key = derived_secret[:self._KEYBYTES]
        return derived_key

    def encrypt(self, *, apu: bytes = b'', apv: bytes = b'') -> (bytes, bytes):
        """
        Encrypt the JWE message, and store the resulting ciphertext and tag.

        :param apu: Agreement PartyUInfo is producer information
            (optional, default: b'').
        :param apv: Agreement PartyVInfo is recipient information
            (optional, default: b'').
        :return: Tuple (ciphertext, tag).
        """
        plaintext = self.get_message_bytes()
        self._header = self._DEFAULT_HEADER.copy()
        ephemeral_key_pair = X25519Jwk(generate=True)
        symmetric_key = self.ecdh_es_key(ephemeral_key_pair, self._jwk,
                                         apu=apu, apv=apv)
        if apu:
            self._header['apu'] = utils.bytes_to_string(apu)
        if apv:
            self._header['apv'] = utils.bytes_to_string(apv)
        self._header['epk'] = ephemeral_key_pair.to_dict(exclude_private=True)
        header_bytes = utils.dict_to_base64(self._header).encode('utf-8')
        self._nonce = nacl.utils.random(self._NONCEBYTES)
        result = self.__class__._aead_encrypt_function(
            plaintext, header_bytes, self._nonce, symmetric_key)
        self._ciphertext, self._tag = result
        return result

    def decrypt(self) -> bytes:
        """
        Decrypts the JWE ciphertext and stores the resulting message.

        :return: Decrypted message.
        """
        ephemeral_key = X25519Jwk(from_dict=self._header['epk'])
        apu = utils.string_to_bytes(self._header.get('apu', ''))
        apv = utils.string_to_bytes(self._header.get('apv', ''))
        symmetric_key = self.ecdh_es_key(self._jwk, ephemeral_key,
                                         apu=apu, apv=apv)
        header_bytes = utils.dict_to_base64(self._header).encode('utf-8')
        result = self.__class__._aead_decrypt_function(
            self._ciphertext, self._tag, header_bytes, self._nonce,
            symmetric_key)
        self._message_bytes = result
        self._message = utils.json_to_dict(result)
        return self._message


class X25519ChaCha20Poly1305Jwe(X25519Jwe):
    """
    X25519 public key encryption with ChaCha20/Poly1305 for the payload.

    Creating an RFC 7518 ECDH-ES key derivation of a suitable key as in RFC
    8037. Payload content is encrypted using ChaCha20/Poly1305 again.

    - https://tools.ietf.org/id/draft-amringer-jose-chacha-00.txt
    - RFC 7539 and RFC 8439
    """

    _DEFAULT_HEADER = OrderedDict([('alg', 'ECDH-ES'),
                                   ('enc', 'C20P'),
                                   ('epk', None)])
    _KEYBYTES = nacl_aead.KEYBYTES
    _NONCEBYTES = nacl_aead.NONCEBYTES
    _aead_encrypt_function = nacl_aead.chacha20poly1305_encrypt
    _aead_decrypt_function = nacl_aead.chacha20poly1305_decrypt


class X25519AES256GCMJwe(X25519Jwe):
    """
    X25519 public key encryption with AES256-GCM for the payload.

    Creating an RFC 7518 ECDH-ES key derivation of a suitable key as in RFC
    8037. Payload content is encrypted using AES256-GCM.

    - RFC 7539 and RFC 8439
    """

    _DEFAULT_HEADER = OrderedDict([('alg', 'ECDH-ES'),
                                   ('enc', 'A256GCM'),
                                   ('epk', None)])
    _KEYBYTES = cryptography_aead.KEYBYTES
    _NONCEBYTES = cryptography_aead.NONCEBYTES
    _aead_encrypt_function = cryptography_aead.aes256gcm_encrypt
    _aead_decrypt_function = cryptography_aead.aes256gcm_decrypt


_CLASS_MAPPING = {
    (ChaCha20Poly1305Jwe._DEFAULT_HEADER['enc'],
     ChaCha20Poly1305Jwe._DEFAULT_HEADER['alg']):
    ChaCha20Poly1305Jwe,
    (AES256GCMJwe._DEFAULT_HEADER['enc'],
     AES256GCMJwe._DEFAULT_HEADER['alg']):
    AES256GCMJwe,
    (UnsecuredJwe._DEFAULT_HEADER.get('enc'),
     UnsecuredJwe._DEFAULT_HEADER['alg']):
    UnsecuredJwe,
    (X25519ChaCha20Poly1305Jwe._DEFAULT_HEADER['enc'],
     X25519ChaCha20Poly1305Jwe._DEFAULT_HEADER['alg']):
    X25519ChaCha20Poly1305Jwe,
    (X25519AES256GCMJwe._DEFAULT_HEADER['enc'],
     X25519AES256GCMJwe._DEFAULT_HEADER['alg']):
    X25519AES256GCMJwe
}



# -*- coding: utf-8 -*-
"""
Wrapper to PyNaCl raw bindings.

Exposes IETF Authenticated Encryption with Associated Data (AEAD) compliant
en-/decryption using ChaCha20/Poly1305 more cleanly.
"""

# Created: 2018-08-02 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

from nacl.bindings import crypto_aead

KEYBYTES = crypto_aead.crypto_aead_chacha20poly1305_ietf_KEYBYTES
NONCEBYTES = crypto_aead.crypto_aead_chacha20poly1305_ietf_NPUBBYTES
ABYTES = crypto_aead.crypto_aead_chacha20poly1305_ietf_ABYTES


def chacha20poly1305_encrypt(message: bytes, aad: bytes,
                             nonce: bytes, key: bytes) -> (bytes, bytes):
    """
    Encrypt the ``message`` using the IETF ratified ChaCha20/Poly1305.

    According to RFC 7539.

    :param message: Message to be encrypted.
    :param aad: Additional authenticated data.
    :param nonce: Nonce.
    :param key: Encryption key.
    :return: Tuple of (ciphertext, tag) bytes.
    """
    result = crypto_aead.crypto_aead_chacha20poly1305_ietf_encrypt(
        message, aad, nonce, key)
    return result[:-ABYTES], result[-ABYTES:]


def chacha20poly1305_decrypt(ciphertext: bytes, tag: bytes, aad: bytes,
                             nonce: bytes, key: bytes) -> bytes:
    """
    Decrypt the ``ciphertext`` using the IETF ratified ChaCha20/Poly1305.

    According to RFC 7539.

    :param ciphertext: Message to be encrypted.
    :param tag: Authentication data (a MAC).
    :param aad: Additional authenticated data.
    :param nonce: Nonce.
    :param key: Encryption key.
    :return: Plaintext bytes.
    :raise: {nacl.exceptions.CryptoError} on an invalid tag.
    """
    return crypto_aead.crypto_aead_chacha20poly1305_ietf_decrypt(
        ciphertext + tag, aad, nonce, key)



# -*- coding: utf-8 -*-
"""
Wrapper to Cryptography hazmat primitives bindings.

Exposes Authenticated Encryption with Associated Data (AEAD) compliant
en-/decryption using AES-GCM more cleanly.
"""

# Created: 2019-03-08 Guy K. Kloss <guy@mysinglesource.io>
#
# (c) 2018-2019 by SingleSource Limited, Auckland, New Zealand
#     http://mysinglesource.io/
#     Apache 2.0 Licence.
#
# This work is licensed under the Apache 2.0 open source licence.
# Terms and conditions apply.
#
# You should have received a copy of the licence along with this
# program.

__author__ = 'Guy K. Kloss <guy@mysinglesource.io>'

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

KEYBYTES = 32
NONCEBYTES = 12
ABYTES = 16


def aes256gcm_encrypt(message: bytes, aad: bytes,
                      nonce: bytes, key: bytes) -> (bytes, bytes):
    """
    Encrypt the ``message`` using AEAD with AES256-GCM.

    :param message: Message to be encrypted.
    :param aad: Additional authenticated data.
    :param nonce: Nonce.
    :param key: Encryption key.
    :return: Tuple of (ciphertext, tag) bytes.
    """
    aesgcm = AESGCM(key)
    result = aesgcm.encrypt(nonce, message, aad)
    return result[:-ABYTES], result[-ABYTES:]


def aes256gcm_decrypt(ciphertext: bytes, tag: bytes, aad: bytes,
                      nonce: bytes, key: bytes) -> bytes:
    """
    Decrypt the ``ciphertext`` using AEAD with AES256-GCM.

    :param ciphertext: Message to be encrypted.
    :param tag: Authentication data (a MAC).
    :param aad: Additional authenticated data.
    :param nonce: Nonce.
    :param key: Encryption key.
    :return: Plaintext bytes.
    :raise: {cryptography.exceptions.InvalidTag} on an invalid tag.
    """
    aesgcm = AESGCM(key)
    return aesgcm.decrypt(nonce, ciphertext + tag, aad)
